import geopandas
import requests
import zipfile
import tempfile

LANDCOVER_URL = 'https://osmdata.openstreetmap.de/download/land-polygons-split-4326.zip'

from .preprocess import create_geometry, gpd_fromlist

def _read_remote_zipped_geofile(url, geofile_path = '', bbox = None):
    """
    Reads a geofile from a remote zipped file.
    
    Args:
        url (str): The URL of the zipped file.
        geofile_path (str): The path of the file inside the zipped file.
        bbox (tuple, optional): The bounding box to filter the data.
        
    Returns:
        geopandas.GeoDataFrame: The GeoDataFrame containing the data.
    """
    
    file_downloaded = requests.get(url)
        
    with tempfile.NamedTemporaryFile(delete=False, mode='wb', prefix='landcover', suffix='.zip') as temp_file:
        temp_file.write(file_downloaded.content)
        zip_path = temp_file.name
        
    with tempfile.TemporaryDirectory() as temp_dir:
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            zip_ref.extractall(temp_dir)
            return geopandas.read_file(f'{temp_dir}/{geofile_path}', bbox=bbox).dissolve()[['geometry']]
    
    
def retrieve_osm(query, user_agent = 'MyApp/1.0 (mymail@gmail.com)'):
    """
    Retrieves OpenStreetMap data for a given query.

    Args:
        query (str): The search query for the desired location.
        user_agent (str, optional): The user agent string to be used in the request headers.
            Defaults to 'MyApp/1.0 (mymail@gmail.com)'.

    Returns:
        geopandas.GeoDataFrame: A GeoDataFrame containing the retrieved OpenStreetMap geometry.
        
    """
    
    endpoint = 'nominatim.openstreetmap.org'
    nominatim_url = f"https://{endpoint}/search.php?q={query}&polygon_geojson=1&format=json"

    response = requests.get(nominatim_url, 
                            headers = { 'User-Agent': user_agent }).json()
    
    features = [{
                    'type': 'Feature',
                    'geometry': create_geometry(res['geojson']),
                    'properties': {'place_name': res['display_name']}
                        
                } for res in response
                ]

    return geopandas.GeoDataFrame.from_features(features, crs='EPSG:4326')

def landcover_from_bbox(bbox):
    """
    Retrieves landcover data for a given bounding box.
    
    Args:
        bbox (tuple): The bounding box to filter the data.
    """
    return _read_remote_zipped_geofile(LANDCOVER_URL, 
                                      geofile_path='land-polygons-split-4326/land_polygons.shp', 
                                      bbox = bbox)
    
def clip_shape_to_landcover(shape, landcover = None):
    """
    Clips a shape to the landcover data.
    
    Args:
        shape (geopandas.GeoDataFrame): The GeoDataFrame containing the shape to be clipped.
        landcover (geopandas.GeoDataFrame, optional): The GeoDataFrame containing the landcover data.
                                                      If None, the landcover data will be retrieved.
    """
    shape_bounds = tuple(gpd_fromlist(shape['geometry']).dissolve().bounds.values[0])
    
    if landcover is None:
        land_area = landcover_from_bbox(bbox = shape_bounds).dissolve()[['geometry']]
    else:
        land_area = landcover.dissolve()[['geometry']]
    
    clipped_shape = shape.clip(land_area)
    
    return clipped_shape


import geopandas
import shapely
import pandas
import numpy
import json

import pygeohash
import pytess
import s2cell
import h3

from .preprocess import gpd_fromlist, create_geometry

UNIVERSAL_CRS = 'EPSG:3857'

def sq_tessellate(base_shape, meters, project_on_crs = None, within = False):
    """
    Function to tessellate a base shape into square polygons.

    Parameters:
    - base_shape: geopandas.GeoDataFrame.
        The base shape to be tessellated.
    - meters: float
        The size of each square polygon in meters.

    Returns:
    - geopandas.GeoDataFrame
        The tessellated polygons as a GeoDataFrame.
    """
    
    if project_on_crs is None:
        project_on_crs = UNIVERSAL_CRS
        
    shape = base_shape.to_crs(project_on_crs).unary_union

    min_x, min_y, max_x, max_y = shape.bounds

    # Find number of square for each side
    x_squares = int(numpy.ceil(numpy.fabs(max_x - min_x) / meters))
    y_squares = int(numpy.ceil(numpy.fabs(min_y - max_y) / meters))

    # Placeholder for the polygon
    polygons = []
    
    for i in range(x_squares):
        x1, x2 = min_x + meters * i, min_x + meters * (i + 1)
        
        for j in range(y_squares):
            y1, y2 = min_y + meters * j, min_y + meters * (j + 1)
            polygon = shapely.geometry.Polygon([(x1, y1), (x1, y2), (x2, y2), (x2, y1)])
            
            if within and polygon.within(shape):
                polygons.append({"geometry": polygon})
            elif not within and polygon.intersects(shape):
                polygons.append({"geometry": polygon})

    squared_tess = geopandas.GeoDataFrame(polygons, crs=project_on_crs).to_crs(base_shape.crs)
        
    return squared_tess

def h3_tessellate(base_shape, resolution, within = False):
    """
    Tessellates a base shape using H3 hexagons.

    Args:
        base_shape (geopandas.GeoDataFrame): The base shape to tessellate in 4326.
        resolution (int): The H3 resolution level.
        within (bool, optional): If True, only H3 hexagons fully contained within the base shape will be returned.
            If False, H3 hexagons intersecting with the base shape will be returned. Defaults to False.

    Returns:
        geopandas.GeoDataFrame: A GeoDataFrame containing the tessellated H3 hexagons.

    """
    
    shape = base_shape.unary_union
    
    if isinstance(shape, shapely.geometry.Polygon):
        shape = shapely.geometry.MultiPolygon([shape])
    
    h3_indexes = set()
    
    for x in shape.geoms:
        
        if not within:
            boundaries = x.boundary
            
            if isinstance(boundaries, shapely.geometry.LineString):
                boundaries = shapely.geometry.MultiLineString([boundaries])
                
            for sub_line in boundaries.geoms:
                for lon, lat in sub_line.coords:
                    h3_indexes.add(h3.geo_to_h3(lat, lon, resolution))
            
        h3_indexes = h3_indexes.union(h3.polyfill_geojson(json.loads(shapely.to_geojson(x)), resolution))
        
    if len(h3_indexes) == 0:
        return geopandas.GeoDataFrame(columns = ['geometry', 'h3-id'], crs = base_shape.crs)
    
    polygons = [{"geometry" : shapely.geometry.Polygon(h3.h3_to_geo_boundary(h3_index, geo_json=True)),
                 'h3-id' : 'h3_'+str(h3_index)} for h3_index in sorted(h3_indexes)]
    
    return geopandas.GeoDataFrame(polygons, crs=base_shape.crs)

def tri_tessellate(base_shape, project_on_crs = None):
    """
    Tessellates a base shape into triangles. 
    Triangles may be used to process irregular polygons using properties of regular polygons.

    Parameters:
    - base_shape: geopandas.GeoDataFrame.
        The base shape to be tessellated.

    Returns:
    - geopandas.GeoDataFrame
        The tessellated triangles as a GeoDataFrame.
    """
    if project_on_crs is None:
        project_on_crs = UNIVERSAL_CRS
        
    shape = base_shape.to_crs(project_on_crs).unary_union
    triangles = shapely.ops.triangulate(shape)
    triangles_gdf = gpd_fromlist(triangles, crs = project_on_crs)
    
    centroid_gdf = geopandas.GeoDataFrame(geometry = triangles_gdf.centroid, crs = project_on_crs)\
                            .sjoin(       gpd_fromlist([shape], crs = project_on_crs), 
                                          how = 'inner', predicate = 'within'
                                    )
    
    triangles_within = triangles_gdf.loc[centroid_gdf.index]
    triangles_within['area'] = triangles_within.geometry.area
    
    return triangles_within.sort_values('area', ascending = False).reset_index(drop = True).to_crs(base_shape.crs)

def _random_point_in_triangle(triangle):
    """
    Generates a random point within a triangle's vertices.

    Parameters:
    - triangle: shapely.geometry.Polygon
        The triangle to generate a random point in.

    Returns:
    - shapely.geometry.Point
        The randomly generated point within the triangle.
    """
    
    v1, v2, v3, _ = numpy.array(list(zip(*triangle.exterior.coords.xy)))
    r1, r2 = numpy.random.random(), numpy.random.random()
    
    pt = v1 * (1.0 - (r1**0.5)) + v2 * (1.0 - r2) * (r1**0.5) + v3 * r2 * (r1**0.5)
    
    return shapely.geometry.Point(pt)

def random_points_in_polygon(base_shape, n_points):
    """
    Generates a set of random points inside a base shape.

    Parameters:
        base_shape (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): The base shape to tessellate.
        n_points (int): the number of points to generate.

    Returns:
    - geopandas.GeoDataFrame
        The GeoDataFrame containing n_points random points within the base shape.
    """
    
    triangles = tri_tessellate(base_shape)
    triangles['probability'] = (triangles['area'] / triangles['area'].sum())
    triangles_selected = numpy.random.choice(triangles['geometry'].values, size = n_points, p = triangles['probability'])
    
    return gpd_fromlist([_random_point_in_triangle(triangle) for triangle in triangles_selected], crs = base_shape.crs)

def vor_tessellate(base_shape, points):
    """
    Perform Voronoi tessellation on a base shape using a set of points.
    
    Args:
        base_shape (geopandas.GeoDataFrame): The base shape to tessellate.
        points (geopandas.GeoDataFrame or int or list or numpy.ndarray): The points to use for tessellation.
            If a GeoDataFrame, it should be in the same coordinate reference system (CRS) as the base shape.
            If an integer, it represents the number of random points to generate within the base shape.
            If a list or numpy.ndarray, it should contain either shapely.geometry.Point objects or pairs of floats
            representing the coordinates of the points.
    
    Returns:
        geopandas.GeoDataFrame: The resulting Voronoi tessellation as a GeoDataFrame.
    """
    
    shape = base_shape.unary_union
    
    if isinstance(points, geopandas.GeoDataFrame):
        points = points.to_crs(base_shape.crs)
    
    elif isinstance(points, int):
        points = random_points_in_polygon(base_shape, points)
    
    elif isinstance(points, list) or isinstance(points, numpy.ndarray):
        if all(isinstance(item, shapely.geometry.Point) for item in points):
            points = gpd_fromlist(points, crs = base_shape.crs)
            
        elif all(len(item) == 2 for item in points) and \
             all(all([isinstance(num, float) for num in item]) for item in points):
                 
            points = gpd_fromlist([shapely.geometry.Point(item) for item in points], crs = base_shape.crs)
    
    earth_boundaries = gpd_fromlist([shapely.geometry.Point(-18000, -9000), 
                                     shapely.geometry.Point(-18000,  9000), 
                                     shapely.geometry.Point( 18000,  9000), 
                                     shapely.geometry.Point( 18000, -9000)], crs = points.crs)
    
    earth_boundaries.index = ['earth_boundaries'] * 4
    
    points = pandas.concat([points, earth_boundaries])
    
    vor = pytess.voronoi(points.geometry.apply(lambda x: (x.x, x.y)).values)
    poly_vor = gpd_fromlist([shapely.geometry.Polygon(polygon) for _, polygon in vor if len(polygon) > 2])
    
    points['point_geometry'] = points.geometry
    
    vor_gdf = poly_vor  .sjoin(points.drop('earth_boundaries'), 
                            how = 'left', 
                            predicate = 'contains')\
                        .rename(columns = {'index_right': 'pt-id'})\
                        .dropna()\
                        .clip(shape)\
                        .reset_index(drop = True)
                        
    vor_gdf['pt-id'] = vor_gdf['pt-id'].astype(int)
    vor_gdf['point_geometry'] = vor_gdf['point_geometry'].apply(lambda x: x.wkt)
    
    return vor_gdf

def _get_s2_cell_corners(s2_id):
    """
    Returns the corners of an S2 cell.
    
    Args:
        s2_id (int): The S2 cell identifier.
        
    Returns:
        numpy.ndarray: An array of tuples containing the corners of the S2 cell.
    """
    
    face, i, j = s2cell.s2_cell_id_to_face_ij(s2_id)
    level = s2cell.cell_id_to_level(s2_id)
    size = 1 << (30 - level)
    
    face_to_lonlat = lambda f, i, j, l : s2cell.cell_id_to_lat_lon(
                                         s2cell.s2cell._s2_face_ij_to_wrapped_cell_id(f, i, j, l))
    
    get_offset = lambda s : [ (s, -s), (s, s), (-s, s), (-s, -s) ]
    
    corners_theoric = numpy.array([tuple(reversed(face_to_lonlat(face, x + i, y + j, level))) for x, y in get_offset(size)])
    corners = numpy.array([tuple(reversed(face_to_lonlat(face, x + i, y + j, level))) for x, y in get_offset(size // 2)])
    
    corners_adjusted = corners - corners.mean(axis = 0) + corners_theoric.mean(axis = 0)
    
    return corners_adjusted

def _s2_polyfill(geojson, level, within = False):
    """
    Polyfills a GeoJSON geometry with S2 cells.
    
    Args:
        geojson (dict): The GeoJSON geometry to polyfill.
        level (int): The S2 level (1 - 30).
        within (bool, optional): If True, only S2 cells fully contained within the base shape will be returned.
            If False, S2 cells intersecting with the base shape will be returned. Defaults to False.
            
    Returns:
        set: A set containing the S2 indexes of the polyfilled geometry.
    """
    
    shape = create_geometry(geojson)
    lon, lat = shape.centroid.coords[0]
    s2_indexes = {s2cell.lat_lon_to_cell_id(lat, lon, level)}
    
    to_check = set(s2_indexes)
    checked = set()

    while to_check:
        current = to_check.pop()
        checked.add(current)

        neighbors = s2cell.cell_id_to_neighbor_cell_ids(current)
        for neighbor in neighbors:
            if neighbor not in checked and neighbor not in to_check:
                neighbor_shape = shapely.geometry.Polygon(_get_s2_cell_corners(neighbor))
            
                joins_with = neighbor_shape.within(shape) if within else neighbor_shape.intersects(shape)
        
                if joins_with:
                    s2_indexes.add(neighbor)
                    to_check.add(neighbor)
                    
    return s2_indexes

def s2_tessellate(base_shape, level, within = False):
    """
    Tessellates a base shape using S2 cells.

    Args:
        base_shape (geopandas.GeoDataFrame): The base shape to tessellate in 4326.
        level (int): The S2 level (1 - 30).
        within (bool, optional): If True, only S2 cells fully contained within the base shape will be returned.
            If False, S2 cells intersecting with the base shape will be returned. Defaults to False.

    Returns:
        geopandas.GeoDataFrame: A GeoDataFrame containing the tessellated S2 cells.

    """
    
    shape = base_shape.unary_union
    
    if isinstance(shape, shapely.geometry.Polygon):
        shape = shapely.geometry.MultiPolygon([shape])
    
    s2_indexes = set()
    
    for x in shape.geoms:
        for s2_cell in set(_s2_polyfill(json.loads(shapely.to_geojson(x)), level, within)):
            s2_indexes.add(s2_cell)
        
    if len(s2_indexes) == 0:
        return geopandas.GeoDataFrame(columns = ['geometry', 's2-id'], crs = base_shape.crs)
    
    polygons = [{"geometry" : shapely.geometry.Polygon(_get_s2_cell_corners(s2_index)),
                 's2-id' : 's2_'+str(s2cell.cell_id_to_token(s2_index))} for s2_index in sorted(s2_indexes)]
    
    return geopandas.GeoDataFrame(polygons, crs=base_shape.crs)

def _get_geohash_corners(geohash):
    """ Get the corners of a geohash."""
    # Decode the geohash to get the bounding box
    lat, lon, lat_err, lon_err = pygeohash.decode_exactly(geohash)

    # Calculate min and max latitudes and longitudes
    min_lat = lat - lat_err
    max_lat = lat + lat_err
    min_lon = lon - lon_err
    max_lon = lon + lon_err

    # Define the corners
    corners = [(min_lon, min_lat),
               (max_lon, min_lat), 
               (max_lon, max_lat),
               (min_lon, max_lat)]

    return corners

def _get_adjacent_geohashes(geohash, direct):
        
        _base32 = '0123456789bcdefghjkmnpqrstuvwxyz'

        neighboring_hashes = { "right"  : { "even" : "bc01fg45238967deuvhjyznpkmstqrwx",
                                            "odd"  : "p0r21436x8zb9dcf5h7kjnmqesgutwvy" },
                               "left"   : { "even" : "238967debc01fg45kmstqrwxuvhjyznp",
                                            "odd"  : "14365h7k9dcfesgujnmqp0r2twvyx8zb" },
                               "top"    : { "even" : "p0r21436x8zb9dcf5h7kjnmqesgutwvy",
                                            "odd"  : "bc01fg45238967deuvhjyznpkmstqrwx" },
                               "bottom" : { "even" : "14365h7k9dcfesgujnmqp0r2twvyx8zb",
                                            "odd"  : "238967debc01fg45kmstqrwxuvhjyznp" } }

        # Used change of parent tile
        borders   = { "right"  : {  "even" : "bcfguvyz", "odd"  : "prxz" },
                      "left"   : {  "even" : "0145hjnp", "odd"  : "028b"  },
                      "top"    : {  "even" : "prxz",     "odd"  : "bcfguvyz" },
                      "bottom" : {  "even" : "028b",     "odd"  : "0145hjnp" } }
        
        if (len(geohash) == 0):
            raise ValueError("The geohash length cannot be 0. Possible when close to poles")
        srcHash = geohash.lower()
        lastChr = srcHash[-1]
        base = srcHash[:-1]

        splitDirection = ['even', 'odd'][len(srcHash)%2]

        if lastChr in borders[direct][splitDirection]:
            base = _get_adjacent_geohashes(base, direct)

        return base + _base32[neighboring_hashes[direct][splitDirection].index(lastChr)]
    
def _get_cornering_geohashes(geohash):
    """Get the geohashes that share a corner with the input geohash."""
    
    # Get the corners of the geohash
    directions = ['right', 'left', 'top', 'bottom']
    cornering_geohashes = {_get_adjacent_geohashes(geohash, direct) for direct in directions}
    
    return cornering_geohashes

def _gh_polyfill(geojson, precision, within = False):
    """Get a list of geohashes within a polygon."""
    
    shape = create_geometry(geojson)
    
    lon, lat = shape.centroid.coords[0]
    gh_indexes = {pygeohash.encode(lat, lon, precision)}
    
    to_check = set(gh_indexes)
    checked = set()

    while to_check:
        current = to_check.pop()
        checked.add(current)

        neighbors = _get_cornering_geohashes(current)
        for neighbor in neighbors:
            if neighbor not in checked and neighbor not in to_check:
                neighbor_shape = shapely.geometry.Polygon(_get_geohash_corners(neighbor))
            
                joins_with = neighbor_shape.within(shape) if within else neighbor_shape.intersects(shape)
        
                if joins_with:
                    gh_indexes.add(neighbor)
                    to_check.add(neighbor)
                    
    return list(gh_indexes)

def gh_tessellate(base_shape, precision, within = False):
    """
    Tessellates a base shape using geohashes.

    Args:
        base_shape (geopandas.GeoDataFrame): The base shape to tessellate in 4326.
        precision (int): The geohash precision (avoid numbers beyond 8, they require too much time).
        within (bool, optional): If True, only geohash cells fully contained within the base shape will be returned.
            If False, geohash cells intersecting with the base shape will be returned. Defaults to False.

    Returns:
        geopandas.GeoDataFrame: A GeoDataFrame containing the tessellated S2 cells.

    """
    
    shape = base_shape.unary_union
    
    if isinstance(shape, shapely.geometry.Polygon):
        shape = shapely.geometry.MultiPolygon([shape])
    
    gh_indexes = set()
    
    for x in shape.geoms:
        for gh_cell in _gh_polyfill(json.loads(shapely.to_geojson(x)), precision, within):
            gh_indexes.add(gh_cell)
        
    if len(gh_indexes) == 0:
        return geopandas.GeoDataFrame(columns = ['geometry', 'gh-id'], crs = base_shape.crs)
    
    polygons = [{"geometry" : shapely.geometry.Polygon(_get_geohash_corners(gh_index)),
                 'gh-id' : 'gh_'+str(gh_index)} for gh_index in sorted(gh_indexes)]
    
    return geopandas.GeoDataFrame(polygons, crs=base_shape.crs)


import numpy
import pandas
import geopandas
import shapely
import haversine

UNIVERSAL_CRS = 'EPSG:3857'

def create_geometry(geom):
    """
    Create a Shapely geometry from a GeoJSON, WKT or WKB representation.
    
    Args:
        geom (dict, str, bytes): The GeoJSON, WKT or WKB representation of the geometry.
        
    Returns:
        shapely.geometry: The Shapely geometry.
    """
    
    if isinstance(geom, dict): 
        return shapely.from_geojson(str(geom).replace("'", '"'))
    elif isinstance(geom, str):
        return shapely.wkt.loads(geom)
    elif isinstance(geom, bytes):
        return shapely.wkb.loads(geom)
    else:
        raise ValueError("Invalid geometry representation.")

def gpd_fromlist(geometries, crs = 'EPSG:4326'):
    """
    Create a GeoDataFrame from a list of geometries.

    Args:
        geometries (list): A list of Shapely geometries.
        crs (str, optional): The coordinate reference system. Default is 'EPSG:4326'.

    Returns:
        geopandas.GeoDataFrame: The GeoDataFrame created from the list of geometries.
    """
    
    valid_geometries = (shapely.geometry.point.Point, 
                        shapely.geometry.linestring.LineString, 
                        shapely.geometry.polygon.Polygon)
    
    # Check if `geometries` is a single geometry
    if isinstance(geometries, valid_geometries):
        geometries = [geometries]
        
    gdf = geopandas.GeoDataFrame(geometry = geometries, crs = crs)
    
    return gdf

def trajectory_detection(df, stop_radius = 0.150, stop_seconds = 300, no_data_seconds = 36000):
    """
    Enrich the events of a user with trajectory stats based on a given set of stop parameters.

    Args:
        df (pandas.DataFrame): DataFrame containing lat, lng, and timestamp columns.
        stop_radius (float): Radius in kilometers to define a stop. Defaults to 0.150 kilometers (150 meters).
        stop_seconds (int): Minimum duration in seconds to consider a stop. Defaults to 300 seconds (5 minutes).
        no_data_seconds (int): Maximum duration in seconds without data to consider it the same trajectory. Defaults to 36000 seconds (10 hours).

    Returns:
        pandas.DataFrame: DataFrame with additional columns: 
        - prev_lat (float): Latitude of the previous point.
        - prev_lng (float): Longitude of the previous point.
        - prev_timestamp (float): Timestamp of the previous point.
        - next_lat (float): Latitude of the next point.
        - next_lng (float): Longitude of the next point.
        - next_timestamp (float): Timestamp of the next point.
        - delta_space (float): Distance in kilometers between the current point and the next point.
        - delta_time (float): Time difference in seconds between the current point and the next point.
        - speed (float): Speed in km/h between the current point and the next point.
        - traj_id (int): Identifier of the trajectory to which the current point belongs.
        - orig_lat (float): Latitude of the origin point of the current trajectory.
        - orig_lng (float): Longitude of the origin point of the current trajectory.
        - dest_lat (float): Latitude of the destination point of the current trajectory.
        - dest_lng (float): Longitude of the destination point of the current trajectory.
        - mean_lat (float): Mean latitude of the trajectory.
        - mean_lng (float): Mean longitude of the trajectory.
        - from_timestamp (int): Timestamp of the point of the trajectory from which the current point comes.
        - start_time (int): Timestamp of the first point of the trajectory.
        - end_time (int): Timestamp of the next point after the trajectory.
        - total_duration (int): Total duration in seconds of the trajectory.
        - pings_in_traj (int): Number of points in the trajectory.
        - is_stop (bool): Indicates whether the trajectory is considered a stop or not on the set conditions.
    """
    
    df = df.sort_values('timestamp').reset_index(drop=True)
    df['lat'] = df['lat'].astype(float)
    df['lng'] = df['lng'].astype(float)
    df['timestamp'] = df['timestamp'].astype(int)
    
    df['prev_lat'] = df['lat'].shift(1)
    df['prev_lng'] = df['lng'].shift(1)
    df['prev_timestamp'] = df['timestamp'].shift(1)
    
    df['next_lat'] = df['lat'].shift(-1)
    df['next_lng'] = df['lng'].shift(-1)
    df['next_timestamp'] = df['timestamp'].shift(-1)

    df['delta_space'] = df.apply(lambda r: haversine.haversine((r['lat'], r['lng']), 
                                                               (r['next_lat'], r['next_lng'])), axis=1)
    
    df['delta_time'] = (df['next_timestamp'] - df['timestamp'])
    
    # in km/h, in case of no delta_time it returns 0 km/h (assuming no movement)
    df['speed'] = (df['delta_space'] / df['delta_time'] * 3600).replace(float('inf'), 0)

    traj_ids = [0]
    waiting_time = 1
    latlngt = df[['lat', 'lng', 'timestamp']].values
    
    for i in range(1, len(df)): 
        lat, lng, t = latlngt[i]
        lat_stop, lng_stop, t_stop = latlngt[traj_ids[-1]]
        prev_t = latlngt[i - 1][2]
        
        if (t - prev_t) > no_data_seconds:
            traj_ids.extend(range(i - waiting_time + 1, i + 1))
            waiting_time = 1
            continue
        
        space_condition = haversine.haversine([lat, lng], [lat_stop, lng_stop]) < stop_radius
        time_condition = (t - t_stop) > stop_seconds
        
        if space_condition and time_condition:
            traj_ids.extend([traj_ids[-1]]*waiting_time)
            waiting_time = 1
                
        elif space_condition:
            waiting_time += 1
        
        else:
            traj_ids.extend(range(i - waiting_time + 1, i + 1))
            waiting_time = 1
            
    traj_ids.extend([traj_ids[-1]]*(waiting_time - 1))
        
    df['traj_id'] = traj_ids
    
    agg_by_traj = df.assign(pings = 1).groupby('traj_id').agg(orig_lat = ('prev_lat', 'first'),
                                                              orig_lng = ('prev_lng', 'first'),
                                                              dest_lat = ('next_lat', 'last'),
                                                              dest_lng = ('next_lng', 'last'),
                                                              mean_lat = ('lat', 'mean'), 
                                                              mean_lng = ('lng', 'mean'), 
                                                              from_timestamp = ('prev_timestamp', 'first'),
                                                              start_time = ('timestamp', 'first'),
                                                              end_time = ('next_timestamp', 'last'),
                                                              total_distance = ('delta_space', 'sum'),
                                                              total_duration = ('delta_time', 'sum'),
                                                              traj_max_speed = ('speed', 'max'),
                                                              pings_in_traj = ('pings', 'sum'))
    
    df = df.set_index('traj_id').join(agg_by_traj).reset_index()

    df['is_stop'] = df['total_duration'] > stop_seconds
    
    return df

def select_blocks(df, filter_id, **kwargs):
    """
    Helper function to filter out strange blocks of data.
    """
    min_radius = kwargs.get('min_radius', None)
    min_radius_condition = df.assign(filtering = True)['filtering']
    
    integrity_speed = kwargs.get('integrity_speed', None)
    sequence_integrity_condition = df.assign(filtering = True)['filtering']
    
    if min_radius is not None:
        min_radius_apply = lambda x: numpy.max(haversine.haversine_vector(x[['lat', 'lng']].values, 
                                                                          x[['lat', 'lng']].values, comb=True)) > min_radius
        
        min_radius_condition = df[filter_id].map(df.groupby(filter_id).apply(min_radius_apply, include_groups=False).to_dict())
        
    if integrity_speed is not None:
        seq_aggregated = df.groupby(filter_id).agg( depart_time = ('timestamp', 'first'),
                                                    dest_time = ('timestamp', 'last'),
                                                    depart_lat = ('lat', 'first'),
                                                    dest_lat = ('lat', 'last'),
                                                    depart_lng = ('lng', 'first'),
                                                    dest_lng = ('lng', 'last'))
        
        seq_aggregated['integrity_speed'] = False
        
        # remove sequentially blocks of points that do not meet the integrity speed
        
        while ~(seq_aggregated['integrity_speed'].all()):
            seq_aggregated['prev_dest_lat'] = seq_aggregated['dest_lat'].shift(1)
            seq_aggregated['prev_dest_lng'] = seq_aggregated['dest_lng'].shift(1)
            
            seq_aggregated['distance_from_previous_seq'] = seq_aggregated.apply(lambda r: haversine.haversine((r['depart_lat'], r['depart_lng']),
                                                                                                            (r['prev_dest_lat'], r['prev_dest_lng'])), axis=1)
            
            seq_aggregated['time_from_previous_seq'] = seq_aggregated['depart_time'] - seq_aggregated['dest_time'].shift(1)
            
            seq_aggregated['speed_from_previous_seq'] = (seq_aggregated['distance_from_previous_seq'] / seq_aggregated['time_from_previous_seq'] * 3600).fillna(0)
            seq_aggregated['integrity_speed'] = seq_aggregated['speed_from_previous_seq'] < integrity_speed
            
            problematic_od = seq_aggregated[~seq_aggregated['integrity_speed']].index
            
            if len(problematic_od) > 0:
                seq_aggregated.drop(problematic_od[0], inplace = True)
                
        sequence_integrity_condition = df[filter_id].isin(seq_aggregated.index)
    
    filtering = min_radius_condition & sequence_integrity_condition
    
    return filtering

def stop_detection(df, max_speed = None):
    """
    Detect stops in a DataFrame based on the 'is_stop' column and an optional maximum speed threshold.

    Args:
        df (pandas.DataFrame): DataFrame containing trajectory data.
        max_speed (float, optional): Maximum speed threshold in km/h. Defaults to None.

    Returns:
        pandas.DataFrame: DataFrame containing the detected stops.
            Columns:
                - stop_id: Identifier of the stop.
                - traj_id: Identifier of the trajectory.
                - mean_lat: Mean latitude of the stop.
                - mean_lng: Mean longitude of the stop.
                - timestamp: Timestamp of the first point of the stop.
                - end_timestamp: Timestamp of the next point after the stop.
                - delta_time: Total duration in seconds of the stop.
                - pings_in_stop: Number of points in the stop.
    """
    df = df.sort_values('timestamp').reset_index(drop=True)
    
    stop_cols = ['traj_id', 'mean_lat', 'mean_lng', 
                 'timestamp', 'end_timestamp', 
                 'delta_time', 'pings_in_stop']
    
    df['consider_stop'] = df['is_stop']
    
    if max_speed is not None:
        df['consider_stop'] = df['consider_stop'] & (df['traj_max_speed'] < max_speed)
    
    stop_df = df[df['consider_stop']].drop(columns = ['timestamp', 'delta_time'])\
                                     .rename(columns = {'start_time' : 'timestamp', 
                                                        'end_time' : 'end_timestamp',
                                                        'total_duration' : 'delta_time',
                                                        'pings_in_traj' : 'pings_in_stop'})[stop_cols]\
                                     .drop_duplicates()\
                                     .reset_index(drop = True)\
                                     .reset_index(names = 'stop_id')
    
    return stop_df
    
def cluster_stops(stop_df, method = 'raw', radius = 150):
    """
    Helper function to cluster stops.
    
    Args:
        stop_df (pandas.DataFrame): DataFrame containing stop data.
        method (str, optional): Method for clustering stops. Options are 'raw' and 'radius'. Defaults to 'radius'.
        radius (float, optional): Radius in meters to cluster stops. Defaults to 150 meters.
    """
    
    stops = stop_df.copy(deep = True)
    
    if method == 'raw':
        stops['loc_lng'] = numpy.round(stops['mean_lng'] * numpy.cos(numpy.radians(stops['mean_lat'])) * 111320 / radius, 0)
        stops['loc_lat'] = numpy.round(stops['mean_lat'] * 111320 / radius, 0)
        stops['cluster_id'] = stops['loc_lat'].astype(str) + '_' + stops['loc_lng'].astype(str)
        
        stop_clusters = stops.groupby('cluster_id').agg({'mean_lat' : 'mean', 'mean_lng' : 'mean'})
        stops['mean_lat'] = stops['cluster_id'].map(stop_clusters['mean_lat'].to_dict())
        stops['mean_lng'] = stops['cluster_id'].map(stop_clusters['mean_lng'].to_dict())
    
    if method == 'radius':
        points = stops.apply(lambda r: shapely.geometry.Point(r['mean_lng'], r['mean_lat']), axis=1)
        stops = geopandas.GeoDataFrame(stops, geometry = points, crs = 'EPSG:4326')
        
        stop_clusters = stops.to_crs(UNIVERSAL_CRS).buffer(radius).to_crs('EPSG:4326').reset_index(name = 'geometry')\
                                .dissolve().reset_index()[['geometry']].explode(index_parts = False).reset_index(drop = True)
        stops = stops.sjoin(stop_clusters, how = 'left', predicate = 'within').rename(columns = {'index_right' : 'cluster_id'})
        stops = stops.set_index('cluster_id').drop(['mean_lat', 'mean_lng'], axis = 1)\
                        .join(stops.groupby('cluster_id').agg({'mean_lat' : 'mean', 'mean_lng' : 'mean'})).reset_index()

    return stops
    
def location_ranking(stop_df, timezone, start_window = '19:00', end_window = '07:00', method = 'most_frequent'):
    """
    Calculate the ranking of location based on stop data.
    The best approach would be to cluster the stops before ranking them.

    Args:
        stop_df (pandas.DataFrame): DataFrame containing stop data.
        timezone (str): Timezone of the stop data.
        start_window (str, optional): Start time of the window period. Defaults to '22:00'.
        end_window (str, optional): End time of the window period. Defaults to '07:00'.
        method (str, optional): Method for ranking locations. 
                                Options are 'most_frequent', 'most_certain', and 'longest'. Defaults to 'most_frequent'.

    Returns:
        pandas.DataFrame: DataFrame containing the ranking of locations based on the specified method.
            Columns:
                - ranking: Position in the ranking according to the criterion.
                - mean_lat: Latitude of the location.
                - mean_lng: Longitude of the location.
                - most_frequent: Number of unique stops in the location.
                - most_certain: Total number of pings in the location.
                - longest: Sum of stop durations in the location.
    """
    methods = ['most_frequent', 'most_certain', 'longest']
    ranking_columns = ['ranking', 'mean_lat', 'mean_lng'] + methods
    
    stops = stop_df.dropna(subset = 'timestamp').sort_values(by='timestamp')
                     
    window_stops = stops['timestamp'].apply(lambda t: pandas.Timestamp(t, unit='s', tz=timezone))
    
    if len(window_stops) == 0:
        return pandas.DataFrame(columns = ranking_columns)
    
    window_traj_id = stops.set_index(pandas.DatetimeIndex(window_stops))\
                          .between_time(start_window, end_window)['traj_id'].values
        
    window_visits = stops[stops['traj_id'].isin(window_traj_id)]
    
    loc_ranking = window_visits .groupby(['mean_lat', 'mean_lng'])\
                                .agg(most_frequent  = ('stop_id', 'nunique'), 
                                     most_certain   = ('pings_in_stop', 'sum'), 
                                     longest        = ('delta_time', 'sum'))\
                                .sort_values(by = [method]+[x for x in methods if x != method], 
                                             ascending = False)\
                                .reset_index()

    return loc_ranking.reset_index(names = 'ranking')

def trip_detection(df, integrity_speed = 250, 
                       integrity_space = 100,
                       integrity_time = 3600,
                       min_max_distance = None,
                       min_pings_in_trip = None,
                       return_one_exec = False):
    """
    Detect trips after trajectory detection.
    
    Args:
        df (pandas.DataFrame): DataFrame containing trajectory data.
        integrity_speed (float): Maximum speed threshold in km/h. Defaults to 250 km/h.
        integrity_space (float): Maximum space threshold in km. Defaults to 100 km.
        integrity_time (float): Maximum time threshold in seconds. Defaults to 3600 seconds (1 hour).
        min_max_distance (float, optional): Minimum distance threshold in km. Defaults to None.
        min_pings_in_trip (int, optional): Minimum number of pings in a trip. Defaults to None.
        return_one_exec (bool, optional): Return the result of the first execution. Defaults to False.
                                          The first execution returns segments of trajectories that are considered trips.
                                          The second execution returns the final trips.               
                                          
    Returns:
        pandas.DataFrame: DataFrame containing the trips' attributes.
            Columns:
                - trip_id: Identifier of the trip.
                - traj_id: Identifier of the trajectory.
                - lat: Latitude of the point.
                - lng: Longitude of the point.
                - timestamp: Timestamp of the point.
                - delta_space: Distance in km between the current point and the next point.
                - delta_time: Time difference in seconds between the current point and the next point.
                - speed: Speed in km/h between the current point and the next point.
                - pings_in_trip: Number of points in the trip.
        
    """
    trip_columns = ['traj_id', 'lat', 'lng', 'timestamp',  
                    'delta_space', 'delta_time', 'speed', 
                    'is_stop', 'pings_in_traj']
    
    df = df[trip_columns].sort_values('timestamp').reset_index(drop=True)
    
    df['break_point'] = (df['speed'] > integrity_speed) | (df['delta_space'] > integrity_space) | (df['delta_time'] > integrity_time)
    
    df['consider_trip'] = ~df['is_stop'] & ~df['break_point']
    
    df.insert(0, 'trip_id', (df['consider_trip'] & (~df['consider_trip'].shift(fill_value=False))).cumsum())
    df['trip_id'] += df['break_point'].cumsum()
    
    trip_df = df[df['consider_trip']].copy(deep = True)
    
    if min_max_distance is not None:
        trip_df['consider_trip_distance'] = select_blocks(trip_df, filter_id = 'trip_id', min_radius = min_max_distance)
        trip_df = trip_df[trip_df['consider_trip_distance']].copy(deep = True)
    
    trip_df['pings_in_traj'] = trip_df['trip_id'].map(trip_df.groupby('trip_id').count()['pings_in_traj'].to_dict())
    
    if min_pings_in_trip is not None:
        trip_df['consider_trip_pings'] = trip_df['pings_in_traj'] > min_pings_in_trip
        trip_df = trip_df[trip_df['consider_trip_pings']].copy(deep = True)
    
    trip_df['integrated_trips'] = select_blocks(trip_df, filter_id = 'trip_id', integrity_speed = integrity_speed)
    trip_df_wattr = trip_df.loc[trip_df['integrated_trips']].reset_index()
    
    trip_df_wattr['next_lat'] = trip_df_wattr['lat'].shift(-1)
    trip_df_wattr['next_lng'] = trip_df_wattr['lng'].shift(-1)
    trip_df_wattr['delta_space'] = trip_df_wattr.apply(lambda r: haversine.haversine((r['lat'], r['lng']),
                                                                                     (r['next_lat'], r['next_lng'])), axis=1)
    trip_df_wattr['next_timestamp'] = trip_df_wattr['timestamp'].shift(-1)
    trip_df_wattr['delta_time'] = trip_df_wattr['next_timestamp'] - trip_df_wattr['timestamp']
    trip_df_wattr['speed'] = trip_df_wattr['delta_space'] / trip_df_wattr['delta_time'] * 3600
    trip_df_wattr['is_stop'] = False 
    
    trip_res = trip_df_wattr[['trip_id']+trip_columns]
    
    if return_one_exec:
        return trip_res.drop(columns = ['is_stop']).rename(columns = {'pings_in_traj' : 'pings_in_trip'})
    
    return trip_detection(trip_res, integrity_speed = integrity_speed, 
                                    integrity_space = integrity_space, 
                                    integrity_time = integrity_time, 
                                    min_max_distance = min_max_distance, 
                                    min_pings_in_trip = min_pings_in_trip, 
                                    return_one_exec = True)
    
def trip_compression(trip_df, return_geometry = False, list_speed = True):
    aggregating_info = {'timestamp' : ('timestamp', 'first'),
                        'end_timestamp' : ('timestamp', 'last'),
                        'lat' : ('lat', 'first'),
                        'next_lat' : ('lat', 'last'),
                        'lng' : ('lng', 'first'),
                        'next_lng' : ('lng', 'last'),
                        'pings_in_trip' : ('pings_in_trip', 'count'),
                        'delta_time' : ('delta_time', lambda x: x[:-1].sum()),
                        'delta_space' : ('delta_space', lambda x: x[:-1].sum())}

    if return_geometry:
        aggregating_info.update({'lat_sequence' : ('lat', list),
                                 'lng_sequence' : ('lng', list)})
    if list_speed:
        aggregating_info.update({'speed_sequence' : ('speed', lambda x: x[:-1].tolist())})

    compressed_trip = trip_df.groupby('trip_id').agg(**aggregating_info)

    if return_geometry:
        compressed_trip['geometry'] = compressed_trip.apply(lambda r: shapely.geometry.LineString(zip(r['lng_sequence'], r['lat_sequence'])), axis=1)
        compressed_trip.drop(['lat_sequence', 'lng_sequence'], axis=1, inplace=True)

    return compressed_trip



import geopandas
import shapely.geometry
from .preprocess import gpd_fromlist, UNIVERSAL_CRS

def get_isochrones(points, max_weight, igraph_graph, graph_weights, concavity = 0.7):
    points = points
    max_weight = max_weight
    igraph_graph = igraph_graph
    graph_weights = graph_weights
    concavity = concavity

    nodes = gpd_fromlist(igraph_graph.vs['geometry'])    
    nearest_nodes = geopandas.sjoin_nearest(points.to_crs(UNIVERSAL_CRS), 
                                            nodes .to_crs(UNIVERSAL_CRS)).to_dict()['index_right']
        
    isochrones = []
    
    for pti, nni in nearest_nodes.items():
        path_tree = igraph_graph.get_shortest_paths(nni, 
                                                    weights = graph_weights, 
                                                    mode='out', 
                                                    output='epath')

        path_tree_weights = [(i, sum(igraph_graph.es[path][graph_weights])) for i, path in enumerate(path_tree)]
        edges = {edge for i, weight in path_tree_weights for edge in path_tree[i] if weight <= max_weight}

        isochrone_edges = shapely.geometry.MultiLineString(igraph_graph.es[list(edges)]['geometry'])
        
        isochrones.append({'point_index': pti, 
                            'point_geometry' : points.loc[pti, 'geometry'], 
                            'nearest_node_index' : nni, 
                            'node_geometry' : nodes.loc[nni, 'geometry'], 
                            'edge_geometry': isochrone_edges,
                            'geometry' : shapely.concave_hull(isochrone_edges, ratio = 1 - concavity)})
                
    return geopandas.GeoDataFrame(isochrones, crs='EPSG:4326', geometry='geometry')
                